/**
 * Created by Ludo Pulles on 2017-12-26.
 */

/*******************************************************************************
 * Sudoku is the object which manages the game visualisation.
 * This part makes sure that the board transitions can be made.
 * When the board is changed, it sends the board to all its controllers.
 ******************************************************************************/

function Sudoku(playerNames, moves) {
	this.playerNames = playerNames;
	this.N = 3;
	this.size = this.N * this.N;

	this.moves = [ "Start" ];
	this.grid = Array(this.size);
	this.hints = Array(this.size);

	if (moves[moves.length - 1].length == 81) {
		this.solution = moves.pop();
	} else {
		this.solution = null;
	}

	for (var i in moves) {
		if (moves[i] == "!") {
			this.moves.push(moves[i]);
		} else {
			this.moves.push({
				row: moves[i].charCodeAt(0) - 65, // A
				col: moves[i].charCodeAt(1) - 97, // a
				num: moves[i].charCodeAt(2) - 48 // 0
			});
		}
	}

	if (this.solution != null) {
		this.moves.push("solution");
	}

	this.isRunning = false;
	this.board = document.getElementById('game-container');
	this.moveTable = document.getElementById('move-table').children[0];

	var C = function(tag, kids, attrs) {
		var x = document.createElement(tag);
		for (i in (kids||[])) x.appendChild(kids[i]);
		if (attrs && attrs['class']) x.className = attrs['class'];
		if (attrs && attrs['text']) x.innerText = attrs['text'];
		return x;
	};

	var head = [ C('th') ], rows = [];
	for (var col = 0; col < this.size; col++) {
		head.push(C('th', [], {text:String.fromCharCode(97 + col)}));
	}

	for (var row = 0; row < this.size; row++) {
		this.grid[row] = Array(this.size);
		this.hints[row] = Array(this.size);
		var currow = [ C('td', [], { text: String.fromCharCode(65 + row), class: 'frow' }) ];
		for (var col = 0; col < this.size; col++) {
			var row_class = row % this.N == 0 ? 'T' : (row % this.N == this.N-1 ? 'B' : '');
			var col_class = col % this.N == 0 ? 'L' : (col % this.N == this.N-1 ? 'R' : '');

			this.hints[row][col] = [];
			for (var i = 0; i < this.N; i++) {
				for (var j = 0; j < this.N; j++) {
					cls = 'Yvis cand row' + i + ' col' + j;
					this.hints[row][col].push(C('div', null, { class: cls, text: i * this.N + j + 1}));
				}
			}
			currow.push(C('td', [ this.grid[row][col] = C('div') ].concat(this.hints[row][col]), { class: row_class + ' ' + col_class }));


		}
		rows.push(C('tr', currow));
	}

	var foot = [ C('td', [], {class:'frow'}) ];

	this.board.appendChild(C('table', [
		C('thead', [C('tr', head)]),
		C('tbody', rows),
		C('tfoot', [C('tr', foot)])
	]));

	this.curMove = 0;

	this.registerEvents();
	this.buildMoveTable(C, moves);
	this.updateSelectedRow();
}

Sudoku.prototype.updateHints = function() {
	if (this.moves[this.curMove] == "solution") {
		var pointer = 0;
		for (var row = 0; row < this.size; row++) {
			for (var col = 0; col < this.size; col++) {
				for (var k = 0; k < this.size; k++) {
					this.hints[row][col][k].className = 'N' + this.hints[row][col][k].className.slice(1);
				}
			}
		}
	} else {
		var inRow = Array(this.size), inCol = Array(this.size), inBox = Array(this.size);
		for (var i = 0; i < this.size; i++) { inRow[i] = Array(this.size); inCol[i] = Array(this.size); inBox[i] = Array(this.size); }

		var to_box = function(r, c) {
			return [0, 0, 0, 3, 3, 3, 6, 6, 6][r] + [0, 0, 0, 1, 1, 1, 2, 2, 2][c];
		};

		for (var i = 1; i <= this.curMove; i++) {
			var move = this.moves[i];
			if (move == "!") continue;
			inRow[move.row][move.num] = true;
			inCol[move.col][move.num] = true;
			inBox[to_box(move.row, move.col)][move.num] = true;
		}

		for (var row = 0; row < this.size; row++) {
			for (var col = 0; col < this.size; col++) {
				for (var num = 1; num <= this.size; num++) {
					var vis = !inRow[row][num] && !inCol[col][num] && !inBox[to_box(row, col)][num] && this.grid[row][col].innerText == '';
					this.hints[row][col][num-1].className = (vis ? 'Y' : 'N') + this.hints[row][col][num-1].className.slice(1);
				}
			}
		}
	}
};

// returns whether the move has been processed
Sudoku.prototype.setMove = function(move) {
	if (move < 0 || move >= this.moves.length) return false;
	if (move == this.curMove) return true;
	var self = this;
	processMove = function(move, mv_nr, add) {
		if (move == "!") return;

		console.log(move, add, mv_nr, self.moves.length);
		if (move == "solution" && add) {
			var pointer = 0;
			for (var row = 0; row < self.size; row++) {
				for (var col = 0; col < self.size; col++, pointer++) {
					if (self.grid[row][col].textContent != "") continue;
					self.grid[row][col].textContent = Number(self.solution[pointer]);
					self.grid[row][col].className = 'sol';
				}
			}
		} else if (move == "solution" && !add) {
			for (var row = 0; row < self.size; row++) {
				for (var col = 0; col < self.size; col++) {
					if (self.grid[row][col].className != '')
						self.grid[row][col].textContent = "";
				}
			}
			for (var i = 1; i <= mv_nr; i++) {
				self.grid[self.moves[i].row][self.moves[i].col].textContent = self.moves[i].num;
			}
		} else {
			self.grid[move.row][move.col].textContent = add ? move.num : "";
		}

		if (move == "!" || move == "solution") return;
	};

	while (this.curMove > move) {
		processMove(this.moves[this.curMove--], this.curMove, false);
	}
	while (this.curMove < move) {
		processMove(this.moves[++this.curMove], this.curMove-1, true);
	}

	this.updateSelectedRow();
	this.updateHints();
	return true;
};

// -----------------------------------------------------------------------------
// Event handlers:
// -----------------------------------------------------------------------------

Sudoku.prototype.setRunning = function(running) {
	if (this.isRunning == running) return;

	this.isRunning = running;
	this.playPause.innerText = running ? 'Pause' : 'Play';

	var self = this, run = function() {
		if (!self.isRunning) return;

		if (self.setMove(self.curMove + 1)) {
			setTimeout(run, 500);
		} else {
			// last move has been reached.
			self.setRunning(false);
		}
	}
	if (this.isRunning) run();
}

Sudoku.prototype.prevMove = function() { this.setRunning(false); this.setMove(this.curMove - 1); };
Sudoku.prototype.nextMove = function() { this.setRunning(false); this.setMove(this.curMove + 1); };

Sudoku.prototype.registerEvents = function() {
	this.playPause = document.getElementById('play-pause');

	var self = this;
	document.getElementById('first').onclick = function() { self.setRunning(false); self.setMove(0); };
	document.getElementById('last').onclick = function() { self.setRunning(false); self.setMove(self.moves.length - 1); };
	document.getElementById('prev').onclick = function() { self.prevMove(); };
	document.getElementById('next').onclick = function() { self.nextMove(); };
	this.playPause.onclick = function() { self.setRunning(!self.isRunning); };
};


Sudoku.prototype.buildMoveTable = function(C, moves) {
	var self = this, players = [ "First", "Second" ];

	var head = [ C('th') ];
	for (var i in players) {
		// cut-off on more than 20 characters in the name
		var playerName = players[i];
		if (this.playerNames) {
			var name = this.playerNames[i];
			if (name.length > 20) name = name.slice(0, 17) + "...";
			playerName = name + " as " + playerName;
		}
		head.push(C('th', [], {text:playerName}));
	}

	var rows = [];
	for (var i = 0; i <= moves.length; i++) {
		var row = [ C('th', [], {text:i+'.'}) ];
		var curp = i == 0 ? 0 : (i - 1) % 2;
		for (var j in players) {
			if (j == curp) row.push(C('td', [], {text: i ? moves[i-1] : "Start"}));
			else row.push(C('td'));
		}
		var tr = C('tr', row, {class: i%2 ? 'zebra' : ''});
		tr.onclick = (function(move) { return function() { self.setMove(move); } })(i);
		rows.push(tr);
	}

	if (this.solution != null) {
		var i = moves.length + 1, row = [
			C('th', [], {text:i+'.'}),
			C('td', [], {text: "unique solution"}),
			C('td', [], {text: "unique solution"}),
		], tr = C('tr', row, {class: i%2 ? 'zebra' : ''});
		tr.onclick = (function(move) { return function() { self.setMove(move); } })(i);
		rows.push(tr);
	}

	this.moveTableBody = C('tbody', rows);
	this.moveTable.onkeydown = function(event) {
		if (event.keyCode == 38 || event.keyCode == 37) {
			self.prevMove();
		} else if (event.keyCode == 40 || event.keyCode == 39) {
			self.nextMove();
		} else {
			return;
		}
		event.preventDefault();
		event.stopImmediatePropagation();
	};

	this.moveTable.appendChild(C('table', [C('thead', [C('tr', head)]), this.moveTableBody]));
};

Sudoku.prototype.updateSelectedRow = function() {
	var lastCurMove = document.getElementById('cur-move');
	var tbl = this.moveTable;

	// reset the id.
	if (lastCurMove) lastCurMove.id = '';

	if (this.curMove >= 0) {
		var curRow = this.moveTableBody.children[this.curMove];
		curRow.id = 'cur-move';

		var topCoord = curRow.offsetTop;
		var bottomCoord = curRow.offsetTop + curRow.offsetHeight - tbl.offsetHeight;
		if (tbl.scrollTop < bottomCoord) {
			tbl.scrollTop = bottomCoord;
		} else if (tbl.scrollTop > topCoord) {
			tbl.scrollTop = topCoord;
		}
	} else {
		tbl.scrollTop = 0;
	}
};
